#include <stdio.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include "common.h"

int stop_received = 0;
pid_t child_pid = 0;
char *pidfile = NULL;
int sleep_time = 1;
int verbose = 0;

void stop_handler(int value) {
	stop_received = 1;
}

void write_pidfile() {
	FILE *f = fopen(pidfile, "w");
	if (f == NULL)
		dies(join("unable to open pidfile", pidfile, NULL));
	if (fprintf(f, "%d\n", child_pid) < 0)
		dies("unable to write to pidfile");
	if (fclose(f) == EOF)
		dies("unable to close pidfile");
}

void start_child(char **args) {
	if (child_pid != 0)
		return;

	child_pid = spawn_child(args, -1, -1, -1, NULL);

	if (pidfile != NULL)
		write_pidfile();

	if (verbose > 1)
		msg(join("child ", args[0], " started, pid ",
			int_s(child_pid), NULL));

	restart_sleep(sleep_time);
}

void usage() {
	die("usage: reliably [-p pidfile] [-v] command ...");
}

int main(int argc, char **argv) {
	setprogname("reliably");

	while (1) {
		char c = getopt(argc, argv, "+p:s:v");
		if (c == -1)
			break;

		switch (c) {
		case 'p':
			pidfile = strdup(optarg);
			break;
		case 's':
			sleep_time = atoi(optarg);
			break;
		case 'v':
			verbose++;
			break;
		default:
			usage();
		}
	}

	if ((argc - optind) < 1)
		usage();

	set_signal_handler(SIGINT, stop_handler);
	set_signal_handler(SIGTERM, stop_handler);

	while (1) {
		int status, pid;

		if (child_pid == 0)
			start_child(&argv[optind]);

		pid = wait(&status);
		if (pid < 0 && errno != ECHILD && errno != EINTR)
			dies("wait failed");

		if (pid == child_pid) {
			if (verbose) {
				if (WIFEXITED(status)) {
					msg(join("child exited, code ",
						int_s(WEXITSTATUS(status)),
						NULL));
				} else if (WIFSIGNALED(status)) {
					msg(join("child exited, signal ",
						int_s(WTERMSIG(status)),
						NULL));
				}
			}
			child_pid = 0;
		}

		if (stop_received) {
			if (verbose)
				msg("stop signal received; exiting");
			kill_child(child_pid, SIGTERM);
			kill_child(child_pid, SIGCONT);
			exit(0);
		}
	}
}

